/*
 *	Functions for analysing i386 architecture systems
*/

#include "../../globals.h"

//#include "magic.h" 


/*
 * Some alchemy:
 */

/*
 * Function that manages mapping from linear to physical address space.
 * 
 */
int i386_lin_read_linear(unsigned long PGD, unsigned long lin_addr, void *out_data, unsigned int priv_block_size){

	//"PGD" is needed, because each process have it's own linear address space. Process 0 means kernel memory.

	//Find in VPT "physical" address, where data should be. (if problem, return 1)
	//Read block_size of that data into out_data (if problem, return 2)
	//return 0

	//unsigned long PDE_addr; //= find_PDE();

	unsigned long physical_offset = lin_addr;
	unsigned long PDE_offset = lin_addr;
	unsigned long PTE_offset = lin_addr;
	unsigned long page_number = lin_addr;
	unsigned long tmp_var = 0;

	//---
	
//	 dbgprint("i386_lin_read_linear on addr: 0x%X,\n",lin_addr); 

	if (PGD == 0) {	//special case
		PGD = image1.VPT_address;
	}//else {	//PGD is linear address of PGD of some process
		
//		PGD=i386_lin_read_linear(0,PGD,&PGD,POI);
//	}

	
	//Kernel linear address space:

	// For extracting addresses I use bitwise and, for extracting offsets I have to shift bits.
        physical_offset = physical_offset & 0x003FFFFF;//  << 10; >>10;//bits 21:0  // 4-Mbyte Page Translation
        PDE_offset = PDE_offset >> 22; //bits 31:22 

//	dbgprint("i386_lin_read_linear: %lx, addPO: 0x%X, PDE_o: 0x%X, PDE_e: 0x%X\n",lin_addr,physical_offset,PDE_offset,image1.VPT_address+PDE_offset*POINTER_SIZE); 

        lseek(inputfd,image1.VPT_address+PDE_offset*POINTER_SIZE,SEEK_SET);// *4 because each entry have 32 bits = 4bytes

        if ( (status=read(inputfd,&page_number,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
                return 35;
        }

//	dbgprint("page_number: 0x%X, pointer size: %d\n",page_number,POINTER_SIZE);

	// TODO: process all necessary special bit, like swapped, etc...

	tmp_var = page_number;
	tmp_var = tmp_var & 0x00000080; //only bit NUMBER 7 (ie. 8th bit)

	if (tmp_var != 0) { //4MB pages are used
		
//		dbgprint("4MB pages used for this page.\n"); 

		page_number = page_number >> 22; //bits 31:22

		//If we wan't to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (MB_4-physical_offset)){ // offset + block_size must be less then 4M 
	//		printf("i386_lin_read_linear: page too small for reading... processing another one.\n");
			//recursion
			
			tmp_var = MB_4-physical_offset;

			i386_lin_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var);

		} else{
			tmp_var=priv_block_size;
		}	

		lseek(inputfd,page_number*MB_4+physical_offset,SEEK_SET); //(4MB page number "large_page_number" + offset)
//		dbgprint("page_number: 0x%X, phys. addr: 0x%X\n",page_number,page_number*MB_4+physical_offset);
		
		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "i386_lin_read_linear: read() in %s: %s.\n",inputfile,errormsg);
	                return 35;
        	}
	} else {   //4KB pages are used
//		dbgprint("4KB pages used for this page.\n");
	
		page_number = page_number & 0xFFFFF000; //bits 31:12 ... This must be done this way, not like commented out example

		PTE_offset = (lin_addr & 0x003FF000) >> 12; //bits 21:12 ... This is offset, so we have to shift bits.//XXX optimalize?
		
		physical_offset = physical_offset & 0x00000FFF;//  << 10; >>10;//bits 11:0
		
	//	page_number = page_number >> 12; //bits 31:12 //PDE entry ??? WTF

//		dbgprint("Searching for PTE in 0x%lx.\n",page_number);

		lseek(inputfd,page_number+PTE_offset*4,SEEK_SET); // *4 because each entry have 32 bits = 4bytes //??? XXX

		if ( (status=read(inputfd,&tmp_var,4)) == -1 ) {
                 	errormsg=strerror(errno);
		        fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
			return 36;
		}
		
//		dbgprint("In PTE I found 0x%lx.\n",tmp_var);

		//tmp_var = tmp_var >> 12; //bits 31:12  //PTE entry
		tmp_var = tmp_var & 0xFFFFF000; // this or this way?

//		dbgprint("Seeking to address  0x%lx.\n",tmp_var+physical_offset);

		lseek(inputfd,tmp_var+physical_offset,SEEK_SET);

		//If we want to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (4096-physical_offset)){ // offset + block_size must be less then 4M
	//		printf("i386_lin_read_linear: page too small for reading... processing another one.\n");
			//recursion
			
			tmp_var = 4096-physical_offset; //wtf?? less then zero?? XXX

			i386_lin_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var); //Process rest of request recursively... beware about effectivity

		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                return 38;
        	}


	}

	return 0;

}

/*
 * Function for verifying if given address contains structure list_head.
 *
 * Verification is done by finding and checking double linked 
 * "list_head". This way is chosen instead of classic 
 * pattern matching, because I believe it is more flexible and 
 * can be used for almost every 2.6 kernel version. (even future;)
 * It does not depend on task_struct definition too much, only 
 * important characteristic is number of list_heads between 
 * "comm" array and list_head "tasks". This is specified by
 * TS_MAGIC_HEADLIST_NUMBER. 
 *
 * Correct list_head can be verified easily:
 * Data at address point to (virtual) address where is stored another
 * list_head. Second pointer in this second head list point back
 * to (virtual) address where we started.
 *
 *  ...   |--------> |---------->  |---------->   ...
 *          <------  | <-----------| <-------- 
 *
 * Note: This will not find processes, that are inactive,
 * has already ended in past, or something similiar. Therefore,
 * if you want to mine as much data as possible, and don't 
 * afraid of too much false positives, use different method.
 * (idetect toolkit, or some greping shell utilities)
 *
 *
 *  i386_lin_verify_list_head (address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int i386_lin_verify_list_head(unsigned long ver_addr){

	unsigned long tmp_addr1,tmp_addr2,tmp_addr3;

	if ( (status=lseek(inputfd,ver_addr,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
                fprintf(stderr, "i386_lin_verify_list_head: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
                return(1);
        }

	if ( (status=read(inputfd,&tmp_addr1,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_lin_verify_list_head: read() in %s: %s.\n",inputfile,errormsg);
                return(2);
        }

	if (tmp_addr1 == 0) {
//		printf("i386_lin_verify_list_head: Zero.\n");
		return 6;
	}

//	printf("Verify_list_head: Co tam je:%X\n",tmp_addr1);

	if ( (status = i386_lin_read_linear(0,tmp_addr1,&tmp_addr2,POINTER_SIZE) == -1 )) {
                printf("i386_lin_verify_list_head: 0x%lx seems is not valid virtual address.\n",tmp_addr1);
                return(3);
        }
	
//	printf("Verify_list_head: Co tam je2:%X\n",tmp_addr2);	

	if ( (status = i386_lin_read_linear(0,tmp_addr2+POINTER_SIZE,&tmp_addr3,POINTER_SIZE) == -1 )) {
                printf("i386_lin_verify_list_head: 0x%lx seems is not valid virtual address.\n",tmp_addr3);
                return(4);
        }
	
	if (tmp_addr1 != tmp_addr3) {
//		printf("i386_lin_verify_list_head: addresses 0x%lx and 0x%lx does not match.\n",tmp_addr1,tmp_addr3);
		return(5);
	}
	
//	printf("		Verify list_head: pointer to: %lx\n",tmp_addr2);

	//We need to verify second pointer of list_head too, there were some false positives
	// same algoritmus, except that "prev" pointer points to (addres - POINTER_SIZE) of reverse pointer.
	

	if ( (status=lseek(inputfd,ver_addr-POINTER_SIZE,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
                fprintf(stderr, "i386_lin_verify_list_head: 2: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
                return(6);
        }

	if ( (status=read(inputfd,&tmp_addr1,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_lin_verify_list_head: 2: read() in %s: %s.\n",inputfile,errormsg);
                return(7);
        }

	if (tmp_addr1 == 0) {
//		printf("i386_lin_verify_list_head: Zero.\n");
		return 8;
	}

//	printf("Verify_list_head: 3:Co tam je:%X\n",tmp_addr1);

	if ( (status = i386_lin_read_linear(0,tmp_addr1+POINTER_SIZE,&tmp_addr2,POINTER_SIZE) == -1 )) {
                printf("i386_lin_verify_list_head: 2: 0x%lx seems is not valid virtual address.\n",tmp_addr1);
                return(9);
        }
	
//	printf("Verify_list_head: 3: Co tam je2:%X\n",tmp_addr2);	

	if ( (status = i386_lin_read_linear(0,tmp_addr2,&tmp_addr3,POINTER_SIZE) == -1 )) {
                printf("i386_lin_verify_list_head: 2: 0x%lx seems is not valid virtual address.\n",tmp_addr3);
                return(10);
        }
	
	if (tmp_addr1 != tmp_addr3) {
//		printf("i386_lin_verify_list_head: 2: addresses 0x%lx and 0x%lx does not match.\n",tmp_addr1,tmp_addr3);
		return(11);
	}
	
//	printf("Verify list_head: 2: pointer to: %lx\n",tmp_addr2);

	return 0;
}


/*
 *  i386_lin_verify_vfsmount (pointer to vfsmount structure)
 *  return 0 on success, otherwise -1
 *
 *  vfsmount structure is "recognized" by simple test:
 *  should contain pointer to string "/dev/".
 *  (char *mnt_name)
 *  This is alone not a strong heuristic, but in combination
 *  with specific chain where structure should be found when trying 
 *  to list open files, it is enough.
 */

int i386_lin_verify_vfsmount(unsigned long ver_addr) {

	int tmp,tmp2;
	unsigned long tmp3;
	char pattern[VFS_MAGIC_PATTERN_LEN];

	for (tmp=1;tmp<VFS_MAGIC_MAX_TRIES;tmp++){
		dbgprint("	i386_lin_verify_vfsmount: %d: %lx ::\n",tmp,ver_addr+(tmp*POINTER_SIZE)); //DBG
		if ( (status = i386_lin_read_linear(0,ver_addr+(tmp*POINTER_SIZE),&tmp3,POINTER_SIZE) == -1 )) { //read pointer to string
                	dbgprint("	i386_lin_verify_vfsmount: 0x%lx seems is not valid virtual address.\n",ver_addr+(tmp*POINTER_SIZE));
                //	return -3;
       		 } else {
			if ( (status = i386_lin_read_linear(0,tmp3,pattern,VFS_MAGIC_PATTERN_LEN) == -1 )) { //read string
	      	        	dbgprint("	i386_lin_verify_vfsmount: 0x%lx seems is not valid virtual address.\n",tmp3);
	        	  //      return -4;
	        	} else {
				tmp2=strncmp(pattern,VFS_MAGIC_PATTERN,VFS_MAGIC_PATTERN_LEN); //compare found string
				if (tmp2 == 0) {	// found "/dev/"
					dbgprint("	i386_lin_verify_vfsmount: Found some \"/dev/\" at offset %d\n",tmp*POINTER_SIZE); //DBG
						return 0;  // Found => It is task_list
					}		
				}
			}
		}

	return -1;	// characterstic string not found => prob. not vfsmount
}



/*
 *  i386_lin_verify_qstr (pointer to "qstr" structure )
 *  return 0 on success, otherwise error
 *
 *  Used for listing of open files per process.
 *    d_name -> name
 *    (qstr)   (char)
 *
 *  This function verifies, if pointer points to qstr structure.
 *  qstr is elemental structure, and we assume, that will not change (there is no reason)
 *  Verification is done by checking, if string that "name" points to is really "len"
 *  bytes long (looking for \0) and "len" is not 0.
 *
 *  This is qstr definition:
 *
 *  struct qstr {
 *          unsigned int hash;
 *          unsigned int len;
 *          const unsigned char *name;
 *  };
 *                          
 *
 */

int i386_lin_verify_qstr(unsigned long ver_addr) {
	
	unsigned long tmp1,tmp2;
//	size_t tmp3;
	char tmp_filename[QS_MAGIC_MAX_FILENAME_SIZE];

	if ( (status = i386_lin_read_linear(0,ver_addr+INTEGER_SIZE,&tmp1,INTEGER_SIZE) == -1 )) { //skip "hash", read string length
                printf("i386_lin_verify_qstr: 0x%lx seems is not valid virtual address.\n",ver_addr+INTEGER_SIZE);
                return(2);
        }

	if ( (status = i386_lin_read_linear(0,ver_addr+(INTEGER_SIZE*2),&tmp2,POINTER_SIZE) == -1 )) { //read pointer to string
                printf("i386_lin_verify_qstr: 0x%lx seems is not valid virtual address.\n",ver_addr+INTEGER_SIZE);
                return(3);
        }

	if (tmp1 > QS_MAGIC_MAX_FILENAME_SIZE) {
		printf("i386_lin_verify_qstr: length too big: %lx.\n",tmp1);
		return -2;
	}

	if ( (status = i386_lin_read_linear(0,tmp2,tmp_filename,tmp1) == -1 )) { //read pointer to string
                printf("i386_lin_verify_qstr: 0x%lx seems is not valid virtual address.\n",ver_addr+INTEGER_SIZE);
                return(4);
        }

//	tmp3=tmp1+1;	//TODO XXX find out why this gives:
//	warning: implicit declaration of function ‘strnlen’
//	tmp3=strnlen((const char *)tmp_filename,tmp3);

	if ( strnlen(tmp_filename,tmp1+1) != (tmp1)) {
		printf("i386_lin_verify_qstr: sizes does not match: %lx vs %d.\n",tmp1,strlen(tmp_filename));
		return -1;
	}


	//try to read whole string
	//if size is ok, return 0

	return 0;
}

/*
 *  i386_lin_verify_dentry (pointer to "dentry" structure )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  Used for listing of open files per process.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  This function determines distances in structure dentry
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int i386_lin_verify_dentry(unsigned long ver_addr) {

	int tmp,tmp2;

	for (tmp=1;tmp<DT_MAGIC_MAX_TRIES;tmp++){
	
		dbgprint("i386_lin_verify_dentry: %d: %lx ::\n",tmp,ver_addr+(tmp*POINTER_SIZE)); //DBG
		tmp2=i386_lin_verify_qstr(ver_addr+(tmp*POINTER_SIZE));
		
		if (tmp2 == 0) {	// found qstr
		
			dbgprint("	i386_lin_verify_dentry: Found some qstr at offset %d\n",tmp*POINTER_SIZE); //DBG

			image1.dt_name_dst=tmp*POINTER_SIZE; //save determined distance

				return 0;  // Found => It is task_list
			}		
		}

	return -1;	// qstr not found
}

/*
 *  i386_lin_determine_up_to_path (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  function that tries to determine distances in structures needed
 *  for listing of open files per process up to f_path structure.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  So this function determines distances in structures files_struct, fdtable, file and path.
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int i386_lin_determine_up_to_path(unsigned long ver_addr) {

	int tmp,tmp2,tmp3,tmp4;
	unsigned long tmpl1,tmpl2,tmpl3,tmpl4;

	for (tmp=1;tmp<TS_MAGIC_FILES_MAX_TRIES;tmp++){	//---Looking for "files" structure
		dbgprint("	i386_lin_determine_up_to_path1: %d: %lx ::\n",tmp,ver_addr+(tmp*POINTER_SIZE)); //DBG
		if ( (status = i386_lin_read_linear(0,ver_addr+(tmp*POINTER_SIZE),&tmpl1,POINTER_SIZE) == -1 )) { //read pointer to "files"
                	printf("	i386_lin_determine_up_to_path: 0x%lx seems is not valid virtual address.\n",ver_addr+(tmp*POINTER_SIZE));
       		 } else {

			for (tmp2=1;tmp2<FS_MAGIC_MAX_TRIES;tmp2++){	//---Looking for "fdt" structure
				dbgprint("	i386_lin_determine_up_to_path2: %d: %lx ::\n",tmp2,tmpl1+(tmp2*POINTER_SIZE)); //DBG
				if ( (status = i386_lin_read_linear(0,tmpl1+(tmp2*POINTER_SIZE),&tmpl2,POINTER_SIZE) == -1 )) { //read pointer to string
		                	printf("	i386_lin_determine_up_to_path: 0x%lx seems is not valid virtual address.\n",tmpl1+(tmp2*POINTER_SIZE));
		       		 } else {

					for (tmp3=1;tmp3<FD_MAGIC_MAX_TRIES;tmp3++){	//---Looking for "fd" structure
						dbgprint("	i386_lin_determine_up_to_path4: %d: %lx ::\n",tmp3,tmpl2+(tmp3*POINTER_SIZE)); //DBG
						if ( (status = i386_lin_read_linear(0,tmpl2+(tmp3*POINTER_SIZE),&tmpl3,POINTER_SIZE) == -1 )) { //read pointer to string
				                	printf("	i386_lin_determine_up_to_path: 0x%lx seems is not valid virtual address.\n",tmpl2+(tmp3*POINTER_SIZE));
				       		 } else {

							for (tmp4=1;tmp4<FL_MAGIC_MAX_TRIES;tmp4++){	//---Looking for "fd" structure
								dbgprint("	i386_lin_determine_up_to_path5: %d: %lx ::\n",tmp4,tmpl3+(tmp4*POINTER_SIZE)); //DBG
								if ( (status = i386_lin_read_linear(0,tmpl3+(tmp4*POINTER_SIZE),&tmpl4,POINTER_SIZE) == -1 )) { //read pointer to string
						                	printf("	i386_lin_determine_up_to_path: 0x%lx seems is not valid virtual address.\n",tmpl3+(tmp4*POINTER_SIZE));
						       		 } else {
			
									tmp4=i386_lin_verify_vfsmount(tmpl4); // Now call verification function	
									if (tmp4 == 0) {	// found !
										dbgprint("	i386_lin_determine_up_to_path6: Found some vfs_struct at offset %lx\n",tmpl3); //DBG
										image1.ts_hl_fs_dst=tmp*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fdt_fd_dst=tmp2*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fl_fpath_dst=tmp3*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fs_vfs_dst=tmp4*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.one_of_paths=tmpl3;		//Linear pointer to one of path structures

										return 0;  // Found => It is task_list
									}
								}	
							}	
						}	
					}		
				}
			}
		}

	}

	return -1;	// Did not found
}

/*
 *  determine_open_files (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  This will definitely not work in 2.4 kernels.
 */

int i386_lin_determine_open_files(unsigned long ver_addr) {

	int tmp,tmp2;
	
	printf(" determine_open_files: 0x%lx.\n",ver_addr);
	tmp=i386_lin_determine_up_to_path(ver_addr+image1.ts_hl_comm_dst);	//"files" is after "comm" field

	if (tmp != 0) {
		printf(" determine_open_files: i386_lin_determine_up_to_path  on 0x%lx returned  %d.\n",ver_addr+image1.ts_hl_comm_dst,tmp);
		return -2;
	}
	for (tmp=1;tmp<FL_MAGIC_MAX_TRIES;tmp++){	//---Looking for "fd" structure
		dbgprint("	determine_open_files2: %d: %lx ::\n",tmp,image1.one_of_paths+(tmp*POINTER_SIZE)); //DBG
		tmp2=i386_lin_verify_dentry(image1.one_of_paths+(tmp*POINTER_SIZE)); // Now call verification function	
	
		if (tmp2 == 0) {	// found !
			dbgprint("	determine_open_files3: Found some dentry at offset %lx\n",image1.one_of_paths+(tmp*POINTER_SIZE)); //DBG
			image1.pt_dentry_dst=tmp*POINTER_SIZE;	//save offset inside path structure (to dentry)
			printf("Determination for listing of open files completed.\n");			
			return 0;  // Found => It is task_list	
		}
	}
		
	return -1;
}

/*
 *  i386_lin_verify_task_struct (pointer to "comm" field of task_struct )
 *  return 0 on success, otherwise -1
 */

int i386_lin_verify_task_struct(unsigned long ver_addr) {

	int tmp,tmp2,tmp3=TS_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6;
	unsigned long tmp4;

	for (tmp=1;tmp<TS_MAGIC_MAX_TRIES;tmp++){
	
//		dbgprint("i386_lin_verify_task_struct: %d: %lx ::\n",tmp,ver_addr-(tmp*POINTER_SIZE)); //DBG
		tmp2=i386_lin_verify_list_head(ver_addr-(tmp*POINTER_SIZE));
		
		if (tmp2 == 0) {	// found some list_head
		
//			dbgprint("	i386_lin_verify_task_struct: Found some list_head at negative offset %d\n",tmp*POINTER_SIZE); //DBG

			image1.ts_headlists[++tmp5]=ver_addr-(tmp*POINTER_SIZE); //save each list_head
//			printf("	i386_lin_verify_task_struct: headlists[%d]:%lx \n",tmp5,image1.ts_headlists[tmp5]);

			tmp++; /* because of there can be (and are) 3 head lists in row, and without this
				  our algorithm would find 5 "list_heads" */

			if (--tmp3 == 0 ) { //maximum list_head's found
					    // this should not happen
				return -7;  // Found => It is task_list
			}		
		}
	}

//	dbgprint("        i386_lin_verify_task_struct: tmp5: %d\n",tmp5);
	
	if (tmp5 >= TS_MAGIC_HEADLIST_NUMBER) { //If this is really task_struct

		tmp2=tmp5-TS_MAGIC_HEADLIST_FRONT_NUMBER;

		//This step is needed, because i386_lin_verify_list_head seems to produce 
		// some false positives, so TS_MAGIC_HEADLIST_FRONT_NUMBER may be not enough accurate.
		tmp3=0; tmp6=0;
		for (tmp=tmp2-TS_MAGIC_HEADLIST_SPREAD;tmp < (tmp2+TS_MAGIC_HEADLIST_SPREAD);tmp++){ 

			tmp3=i386_lin_list_head_size(image1.ts_headlists[tmp]);

			if (tmp3 > tmp6) {// more and not loop (loop < 0)
				tmp6=tmp3;			//remember local maximum
				image1.ts_headlists_number=tmp; //save which list_head is right one
			}
		}

		// So at this point we know position of right list_head

//		dbgprint("	i386_lin_verify_task_struct: Found tasks at negative offset %d\n",tmp*POINTER_SIZE); //DBG
		image1.ts_hl_comm_dst=ver_addr-image1.ts_headlists[image1.ts_headlists_number]+POINTER_SIZE; //set distance from headlist to comm for this kernel.
		
		//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
		if ( (status=lseek(inputfd,image1.ts_headlists[image1.ts_headlists_number],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "i386_lin_verify_task_struct: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "i386_lin_verify_list_head: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
	
		image1.ts_one_of_lists=tmp4;
		return 0;  // Found => It is task_list
	}
	//otherwise it is not task_struct
	return -1;
}

/*
 *
 *  Interesting fields in "module" are just near. This should not change (no reason),
 *  so this function is much simpler then i386_lin_verify_task_struct .
 *
 *  i386_lin_verify_module (pointer to "name" field of module )
 *  return distance between name and list_head in bytes on success, otherwise < 0
 *
 *  
 */

int i386_lin_verify_module(unsigned long ver_addr) {

	int tmp2,tmp3;
	unsigned long tmp4;

//	dbgprint("verify_module: %lx: %lx ::\n",ver_addr,ver_addr-(1*POINTER_SIZE)); //DBG
	tmp2=i386_lin_verify_list_head(ver_addr-(1*POINTER_SIZE)); /// XXX - and +
		
	if (tmp2 == 0) {	// found some list_head
		//dbgprint("	verify_module: Found some list_head at negative offset %d\n",1*POINTER_SIZE); //DBG
		image1.md_headlists[0]=ver_addr-(1*POINTER_SIZE); //save list_head
		//dbgprint("	verify_module: headlists[%d]:%lx \n",tmp2,image1.ts_headlists[0]);
	
		 //If this is really module

		tmp3=i386_lin_list_head_size(image1.ts_headlists[0]);
		//dbgprint("        verify_module: tmp3 (size): %d of %lx\n",tmp3,image1.md_headlists[0]);
		if (tmp3 > 0) {// more and not loop (loop < 0)
			image1.md_headlists_number=0; //save which list_head is right one
		}
		

		// So at this point we know position of right list_head

		printf("	i386_lin_verify_module: Found module struct at negative offset %d\n",1*POINTER_SIZE+INTEGER_SIZE); // + integer size because of enum
		image1.md_hl_name_dst=ver_addr-image1.md_headlists[0]+POINTER_SIZE; //set distance from headlist to name for this kernel.
		
		//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
		if ( (status=lseek(inputfd,image1.md_headlists[0],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "i386_lin_verify_module: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "i386_lin_verify_module: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
		dbgprint("i386_lin_verify_module: image1.md_one_of_lists: 0x%x",tmp4);	
		image1.md_one_of_lists=tmp4; //hm, dirty trick, this should be in determine module
		return 0;  // Found => It is module
	}
	//otherwise it is not module
	return -1;
}

/*int i386_lin_verify_module(unsigned long ver_addr) {

	int tmp,tmp2,tmp3=MD_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6;
	unsigned long tmp4;

	for (tmp=1;tmp<MD_MAGIC_MAX_TRIES;tmp++){
	
		dbgprint("verify_module: %d: %lx ::\n",tmp,ver_addr-(tmp*POINTER_SIZE)); //DBG
		tmp2=i386_lin_verify_list_head(ver_addr-(tmp*POINTER_SIZE)); /// XXX - and +
		
		if (tmp2 == 0) {	// found some list_head
		
			printf("	verify_module: Found some list_head at negative offset %d\n",tmp*POINTER_SIZE); //DBG

			image1.md_headlists[++tmp5]=ver_addr-(tmp*POINTER_SIZE); //save each list_head
			printf("	verify_module: headlists[%d]:%lx \n",tmp5,image1.ts_headlists[tmp5]);

			tmp++; // because of there can be (and are) 3 head lists in row, and without this
				  our algorithm would find 5 "list_heads" 

			if (--tmp3 == 0 ) { //maximum list_head's found
					    // this should not happen
				printf("	verify_module: max reached. Not ideal.\n");
				break; //don't give up
//				return -7;  // Found => It is task_list
			}		
		}
	}

	dbgprint("        verify_module: tmp5: %d\n",tmp5);
	
	if (tmp5 >= 1) { //If this is really module

		//This step is needed, because i386_lin_verify_list_head seems to produce 
		// some false positives, so TS_MAGIC_HEADLIST_FRONT_NUMBER may be not enough accurate.
		tmp3=0; tmp6=0;
		for (tmp=0;tmp <= tmp5;tmp++){ //or tmp5+1?

			tmp3=i386_lin_list_head_size(image1.ts_headlists[tmp]);
			dbgprint("        verify_module: tmp3 (size): %d of %lx\n",tmp3,image1.md_headlists[tmp]);
			if (tmp3 > tmp6) {// more and not loop (loop < 0)
				tmp6=tmp3;			//remember local maximum
				image1.md_headlists_number=tmp; //save which list_head is right one
			}
		}

		// So at this point we know position of right list_head

		printf("	verify_module: Found module at negative offset %d\n",tmp*POINTER_SIZE); //DBG
		image1.md_hl_name_dst=ver_addr-image1.md_headlists[image1.md_headlists_number]+POINTER_SIZE; //set distance from headlist to name for this kernel.
		
		//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
		if ( (status=lseek(inputfd,image1.md_headlists[image1.md_headlists_number],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "verify_module: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "verify_module: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
	
		image1.md_one_of_lists=tmp4;
		return 0;  // Found => It is module
	}
	//otherwise it is not module
	return -1;
}
*/

/*
int i386_lin_verify_module(unsigned long ver_addr) {

	int tmp,tmp2,tmp3=TS_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6;
	unsigned long tmp4;

	tmp2=i386_lin_verify_list_head(ver_addr-POINTER_SIZE);	//This should not change.... ok for now
	
	if (tmp2 == 0) {	// found some list_head
		
		image1.md_one_of_lists=ver_addr-POINTER_SIZE;
		printf("	verify_module: Found some list_head. %d\n"); //DBG

	} else {	
		return -1;	// Not module
	}

	//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
	if ( (status=lseek(inputfd,image1.md_one_of_lists,SEEK_SET)) == -1 ) {
                errormsg = strerror(errno);
                fprintf(stderr, "i386_lin_verify_task_struct: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
                return(-2);
        }

        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_lin_verify_list_head: read() in %s: %s.\n",inputfile,errormsg);
                return(-3);
        }

	image1.md_one_of_lists=tmp4;
	return 0;  // Found => It is module
}
*/

/*
 * Function for verifying if given address contains VPT 
 * (resp. PaGeDirectory) of some process.
 *
 * Verification is done by simple checking
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, as userspace PGD are 
 * cloned from kernel PGD, but it is just heuristic, 
 * and can be wrong.
 *
 *  i386_lin_user_i386_lin_VPT_test (linear address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */


int i386_lin_user_i386_lin_VPT_test(unsigned long my_addr){

	int mi=0;

	if ( (status=i386_lin_read_linear(0,my_addr,tmp_small_page,VPT_MAGIC_BS)) != 0 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_lin_user_i386_lin_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
//	printf("i386_lin_user_i386_lin_VPT_test: %lx\n",my_addr);

	mi=VPT_MAGIC_ZEROES; 

	//chceck if structure seems to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information

//	printf("dbg: %2X %2X %2X\n",tmp_small_page[mi]-0xFFFFFF00,tmp_small_page[mi+4]-0xFFFFFF00,tmp_small_page[mi+8]-0xFFFFFF00); 
//	printf("user_vpt_test: dbg :i: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); 

	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE] & 0x6f)) {return 2;}
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE*2] & 0x6f)) {return 3;}
	if (tmp_small_page[mi] == 0) {return 4;}
	if (tmp_small_page[mi] == 0xFFFFFFFF) {return 5;}
//	if ((tmp_small_page[mi] & 0x000000FF) != 0x000000E3) {return 6;} //XXX Just try XXX

//	printf("user_vpt_test: dbg2: :%X,%d,%c\n",tmp_small_page[mi]),tmp_small_page[mi],tmp_small_page[mi];

	return 0; //all OK			
}

/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * Verification is done by simple checking
 * if first VPT_MAGIC_ZEROES bytes contain only zeroes and 
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, but it is just heuristic, 
 * and can be wrong.
 *
 *  i386_lin_VPT_test (physical address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int i386_lin_VPT_test(int my_addr){

	int mi=0;

	if ( (status=read(inputfd,&tmp_small_page,VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_lin_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < VPT_MAGIC_ZEROES;mi++) { //first 
		if (tmp_small_page[mi] != 0) {
			return 1;
		}
	}
	
	//check if structure seems to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information

	//0x6f ~ 11011111 
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE] & 0x6f)) {return 2;}
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE*2] & 0x6f)) {return 3;}
	if (tmp_small_page[mi] == 0) {return 4;}
	if (tmp_small_page[mi] == 0xFFFFFFFF) {return 5;}

//	dbgprint("	vpt_test: dbg :i: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); //DBG
	
	return 0; //all OK			
}

/*
 * Function for finding VPT of given process 
 * (resp. PaGeDirectory).
 *
 * This function is assumed to be called only once.
 * When distances used in analyzed image are known,
 * it is faster and safer to use them directly.
 *
 * Function tries address around values that was used in 
 * available linux 2.6 kernel, and tries, if they match definition.
 * Success is not guaranteed, but it was working nice even for kernels
 * with random debugging symbols compiled in.
 *
 * i386_lin_find_process_VPT (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in glibal structure image1.
 *
 * */

int i386_lin_find_process_VPT(unsigned long my_addr){

	int mi=0,tmp_mi;
	unsigned long tmp2,tmp3;

//	printf("i386_lin_find_process_VPT: read: %lx,%d, dst:%d\n",my_addr+TS_MAGIC_HL_MM_DST,POINTER_SIZE,TS_MAGIC_HL_MM_DST);

	for (tmp_mi=0;tmp_mi < TS_MAGIC_HL_MM_DST*2;tmp_mi+=POINTER_SIZE) { //guessing "ts_hl_mm_dst"

		if ( (status=i386_lin_read_linear(0,my_addr+tmp_mi,&tmp2,POINTER_SIZE)) != 0 ) { //first read pointer
	                errormsg=strerror(errno);
	                fprintf(stderr, "i386_lin_find_process_VPT: read() in %s: %s.\n",inputfile,errormsg);
	        } else { //read OK

			for (mi=0;mi < MM_MAGIC_PGD_DST*2;mi+=POINTER_SIZE) { //guessing "mm_pgd_dst"
		
				if ( (status=i386_lin_read_linear(0,tmp2+mi,&tmp3,POINTER_SIZE)) != 0 ) {
		        	        errormsg=strerror(errno);
		               		fprintf(stderr, "i386_lin_find_process_VPT: read() in %s: %s.\n",inputfile,errormsg);
		               		//return 99; //XXX I should handle this in better way XXX
		        	}
				
				if ( i386_lin_user_i386_lin_VPT_test(tmp3) == 0) { //if this pointer points to VPT (resp. PGD)	
					image1.mm_pgd_dst=mi; //then set distance we found (inside mm_struct from start to pgd)
					image1.ts_hl_mm_dst=tmp_mi; // and set distance (inside task_struct from hl "tasks" to mm) 
					printf("Found userspace PGD at 0x%lx and determined distances: mm_pgd: %d, hl_mm: %d\n",tmp3,mi,tmp_mi);
					return 0;
				}
			}
		}
	}
	
	return 1; //did not found VPT(resp. PGD)
}

/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 *
 * determine ()
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure image1.
 *
 * */

int i386_lin_determine(){
	
	int addr=VPT_MAGIC_ADDR;

	//TODO
	
	//Get some ideas, how to do it...
	//Get as much dumps as possible
	//Make some observations..
	
	//For now hard-coded: (linux 2.6.23, i386)

	image1.guest_block_size=MB_4; //4M  //size of block in analyzed image 
	//cha-chaa, on real system,we can have 2 block sizes at once!
//        image1.architecture="i386";  // hw architecture of analyzed image
//        image1.operating_system="Linux 2.6.22-3";      // operation system if analyzed image
//        image1.ram_size=1048576; //1G   // size of image in kilobytes //!!! fake !!!

// let's find VPT_address:
	
	lseek(inputfd,VPT_MAGIC_ADDR,SEEK_SET); 
	
	while (i386_lin_VPT_test(addr) != 0) {

		addr += VPT_MAGIC_BS; // 4KB
		lseek(inputfd,addr,SEEK_SET);
		
		if (addr > ( VPT_MAGIC_ADDR + VPT_MAGIC_BS * VPT_MAGIC_COUNT) ) { 
			//we have tried too many pages, so something is wrong probably
			//further search will be useless
			printf("VPT_address not found, terminating at: 0x%X\n",addr);
			return 1;
		}
	}
	
	image1.VPT_address=addr; // found one! :)
	printf("VPT_address found at: 0x%lX\n",image1.VPT_address);
	
	return 0;  // 0 if all OK, otherwise error
}

/* Only printf processes for now.
 * listhead is address of list_head of one of the processes
 * distance is in bytes from given listhead structure to comm (char[16]) field
 */

int i386_lin_list_processes(unsigned long listhead,int distance){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[16];
	int tmp_help=1;

	printf("Listing processes: \n(linear addresses of kernel, usually substitute 0xC0000000 to get position in dump file.)\n");
	tmp_list = listhead;

	while (tmp_help) { //This is way "down" in list_head
	
		i386_lin_read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No guaranteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("Listing processes done.\n");
//					tmp_help=0;
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > 65535) { //there cannot be more than 65K processes on Linux (ehm;)
//				tmp_help =0;
				printf("	list_processes: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			i386_lin_read_linear(0,tmp_list+distance,tmp_comm,16);
			printf("	process address: %lx, name: %s\n",tmp_list,tmp_comm); //XXX
		}	
	}

	return -5; //Something is wrong
}

/*
	We do not search for in init directly, because it is too common string in memory => takes too long
	This is almost clone of list_processes
*/

int i386_lin_find_userspace_process(){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[16];
	int tmp_help=1;

	printf("	i386_lin_find_userspace_process: Looking for userspace process \"%s\".\n",MAGIC_USER_PROCESS); 
	
	tmp_list = image1.ts_one_of_lists; //This is already from list_head, but not userspace

	while (tmp_help) { //This is way "down" in list_head
	
		i386_lin_read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
//		printf("	i386_lin_find_userspace_process: tmp:%lx, tmp2:%lx.\n",tmp_list,tmp_list2); 

		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) { //List is empty, or invalid linear address was provided 
			debug();
			printf("fusp: strange,1: %lx,2: %lx\n",tmp_list,tmp_list2);
			tmp_help=0;
		} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No guaranteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("	i386_lin_find_userspace_process: Not found \"%s\", something is wrong.\n",MAGIC_USER_PROCESS);
//					tmp_help=0;
					return -1;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > 65535) { //there cannot be more than 65K processes on Linux (ehm;)
//				tmp_help =0;
				printf("	i386_lin_find_userspace_process: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			i386_lin_read_linear(0,tmp_list+image1.ts_hl_comm_dst,tmp_comm,16);
			if ( (strncmp(tmp_comm, MAGIC_USER_PROCESS, strlen(MAGIC_USER_PROCESS))) == 0 ) {
				printf("Found userspace process \"%s\" at 0x%lx\n",MAGIC_USER_PROCESS,tmp_list); //XXX
				image1.ts_one_of_lists=tmp_list;
				return 0;
			}
		}	
	}

	return -5; //Something is wrong, list is empty or invalid
}

/* Another clone of list_processes
 * Only number of members of list_head
 *
 * input: Physical address of list_head
 * output: Number of members, -3 if infinite loop detected
 */

int i386_lin_list_head_size(unsigned long listhead){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	int tmp_help=1;

	if ( (status=lseek(inputfd,listhead,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
	        fprintf(stderr, "i386_lin_list_head_size: lseek() to 0x%lx failed because of %s.\n",listhead,errormsg);
	        return(-1);
	}
	
	if ( (status=read(inputfd,&tmp_list,POINTER_SIZE)) == -1 ) {
	        errormsg=strerror(errno);
	        fprintf(stderr, "i386_lin_list_head_size: read() in %s: %s.\n",inputfile,errormsg);
	        return(-2);
	}
//	printf("	i386_lin_list_head_size: %lx\n",listhead);

	while (tmp_help) { //This is way "down" in list_head
	
		i386_lin_read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No garanteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
//					printf("	i386_lin_list_head_size: counter:%d.\n",counter);
//					tmp_help=0;
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > 65535) { //there cannot be more than 65K processes on Linux (ehm;)
//				tmp_help =0;
				printf("	i386_lin_list_head_size: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
//			i386_lin_read_linear(0,tmp_list+distance,tmp_comm,16);
//			printf("	address: %lx, name: %s\n",tmp_list,tmp_comm); //XXX
		}	
	}

	return -5; //Something is wrong
}

/* Only printf modules for now.
 * listhead is address of list_head of one of the modules (not linear address, but address in physical dump file)
 * distance is in bytes from given listhead structure to comm (char[16]) field
 */


int i386_lin_list_modules(unsigned long listhead,int distance){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[64]; //according to module.h, with  reserve
	int tmp_help=1;

	printf("Listing modules:\n(linear addresses of kernel, usually substitute 0xC0000000 to get position in dump file.)\n");
	tmp_list = listhead;

	while (tmp_help) { //This is way "down" in list_head
	
		i386_lin_read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same module twice. No guaranteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("Listing modules done.\n");
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of modules found
			if (loop_detection++ > 65535) { 
				//there cannot be more than 65K modules on Linux (ehm;)
				// XXX check XXX
				printf("	list_modules: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			i386_lin_read_linear(0,tmp_list+distance,tmp_comm,64);
			printf("	module address: %lx, name: %s\n",tmp_list,tmp_comm); //XXX
		}	
	}

	return -5; //Something is wrong
}

/*
 * Function that tries to determine linux-specific information 
 * about analyzed image.
 *
 * Some process and then some userland process are found,
 * and distance between important fields are determined.
 *
 * Information about "module" structure are determined separately,
 * because module support is only optional.
 *
 * linux_determine ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int i386_lin_linux_determine(){
	
	int tmp;
	unsigned long tmp2=0;
	
	printf("	linux_determine: Looking for process \"%s\".\n",MAGIC_PROCESS);

 	if (search_init(inputfile,f,MAGIC_PROCESS) != 0 ) { //This seems to work, even if it is not optimal
		printf("linux_determine: search_init retruned error. Aborting.\n");
		return -1;
	}

	
	while ((tmp2=search_next())!=0){
//		tmp2-=POINTER_SIZE; // because of way how is i386_lin_verify_task_struct written.. maybe I should change this XXX
//		printf("	linux_determine: Trying to verify: %lx as task_struct\n",tmp2);
		if (i386_lin_verify_task_struct(tmp2) == 0) { //we have found valid process
			printf("Found valid process at 0x%lx.\n",tmp2);
			break;
		}
	}

	if (tmp2 == 0) { //Found nothing
		printf("	linux_determine: Cannot find process %s, exiting.",MAGIC_PROCESS); //XXX
		return 2;
	}
//	printf("	linux_determine: tmp2: %lx, image1.ts_one_of_lists: %lx\n",tmp2,image1.ts_one_of_lists);//DBG

	// Find some userspace process
	if ( (tmp=i386_lin_find_userspace_process()) !=0) {
		printf("	linux_determine: i386_lin_find_userspace_process returned: %d, terminating.\n",tmp);
		return 3;
	}
	
//	printf("	linux_determine: 2: tmp2: %lx, image1.ts_one_of_lists: %lx\n",tmp2,image1.ts_one_of_lists);//DBG
	// determnine VPT distances
	if ( (tmp=i386_lin_find_process_VPT(image1.ts_one_of_lists)) !=0) {
		printf("	linux_determine: i386_lin_find_process_VPT returned: %d, terminating.\n",tmp);
		return 4;
	}
      
	printf("Determination process complete.\n");//DBG

	return 0;
}

/*
 * Function that tries to determine linux module-specific 
 * information about analyzed image.
 *
 * Some module is found,
 * and distance between important fields are determined.
 *
 * This is done separately,
 * because module support is only optional, and is not 
 * critical for another analyze.
 *
 * linux_determine_module ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int i386_lin_linux_determine_module(){
	
//	int tmp;
	unsigned long tmp2=0;
	
	printf("	linux_determine_module: Looking for module \"%s\".\n",MAGIC_MODULE);

 	if (search_init(inputfile,f,MAGIC_MODULE) != 0 ) { //This seems to work, even if it is not optimal
		printf("linux_determine: search_init retruned error. Aborting.\n");
		return -1;
	}

	
	while ((tmp2=search_next())!=0){
//		tmp2-=POINTER_SIZE; // because of way how is i386_lin_verify_task_struct written.. maybe I should change this XXX
//		printf("	linux_determine: Trying to verify: %lx as task_struct\n",tmp2);
		if (i386_lin_verify_module(tmp2) == 0) { //we have found valid module

			printf("Found valid module at 0x%lx.\n",tmp2);
			break;
		}
	}

	if (tmp2 == 0) { //Found nothing
		printf("	linux_determine: Cannot find module %s, exiting.\n",MAGIC_MODULE); //XXX
		return 2;
	}
//	printf("	linux_determine: tmp2: %lx, image1.ts_one_of_lists: %lx\n",tmp2,image1.ts_one_of_lists);//DBG

	printf("Determination process of module structure complete.\n");

	return 0;
}

