/* And heres another throw -- binfsrch.c by CBF */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

/* The difference between a binary and a text file, on read,
is the conversion of end-of-line delimiters. What those
delimiters are does not affect the action. In some cases
the presence of 0x1a EOF markers (MsDos) does.

This is a version of Knuth-Morris-Pratt algorithm. The
point of using this is to avoid any backtracking in file
reading, and thus avoiding any use of buffer arrays.
*/

unsigned long chrcount=0; /* count of input chars, zeroed */
char marker[31];
int *next;
int lgh;
int ch;
int items; /* count of markers found */
FILE *f;

/* --------------------- */

/* Almost straight out of Sedgewick */
/* The next array indicates what index in id should next be
compared to the current char. Once the (lgh - 1)th char
has been successfully compared, the id has been found.
The array is formed by comparing id to itself. */

void initnext(int *next, const char *id, int lgh)
{
	int i, j;
	
	assert(lgh > 0);
	next[0] = -1; i = 0; j = -1;
	while (i < lgh) {
		while ((j >= 0) && (id[i] != id[j])) j = next[j];
		i++; j++;
		next[i] = j;
	}
	#ifdef DEBUG
	for (i = 0; i <= lgh; i++)
		printf("id[%d] = '%c' next[%d] = %d\n", i, id[i], i, next[i]);
	#endif
} /* initnext */

/* --------------------- */

/* reads f without rewinding until either EOF or *marker
has been found. Returns EOF if not found. At exit the
last matching char has been read, and no further. */
int kmpffind(const char *marker, int lgh, int *next, FILE *f)
{
	int j; /* char position in marker to check */
	int ch; /* current char */

	assert(lgh > 0);
	j = 0;
	while ((j < lgh) && (EOF != (ch = getc(f)))) {
		chrcount++;
		while ((j >= 0) && (ch != marker[j])) j = next[j];
		j++;
	}
	return ch;
} /* kmpffind */

/* --------------------- */

/* Function that initialize search state.
 * (reopen file and null counter)
*/

int search_init(char *inputfile2,FILE *f2,const char *marker2) {

//	printf("search_init: %s, %s\n",inputfile2,marker); /// DBG
	f=f2;	//Make a local copy of all variables
	memcpy(marker,marker2,strlen(marker2)); //XXX
	chrcount=0;

	if (!(f = fopen(inputfile2, "rb"))) {
		printf("Can't open %s\n", inputfile2);
		return -1;
	}

	lgh = strlen(marker);
	if (!(next = malloc(1 + lgh * sizeof *next))) {
		puts("No memory");
		return -2;
	}	
//	printf("search_init: lgh:%d\n", lgh);	//DBG
	
	initnext(next, marker, lgh);
	items = 0;
	return 0;
}

unsigned long search_next()
{	
//	printf("search_next: %d %s : %x\n", items, marker,chrcount-strlen(marker)); //DBG
	if (EOF == kmpffind(marker, lgh, next, f)) { //find next
		free(next);
		return 0;
	}
	return chrcount-strlen(marker);

} /* binfsrch */




